/*
 * Copyright (C) 1997 and 1998 WIDE Project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by WIDE Project and
 *	its contributors.
 * 4. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
/*
 * $Id: draw.c,v 1.33.2.113 1998/01/16 14:27:31 kazu Exp $
 */

#include "mgp.h"

/* state associated with the window - how should we treat this? */
static struct ctrl *bg_image_ctl = NULL;
static struct ctrl *bg_image_prev = NULL;
static Image *bg_image = NULL;
static XImageInfo *bg_ximageinfo = NULL;

static u_short kinsokutable[] = {
	0x2121, 0x2122, 0x2123, 0x2124, 0x2125, 0x2126, 0x2127, 0x2128,
	0x2129, 0x212a, 0x212b, 0x212c, 0x212d, 0x212e, 0x212f, 0x2130,
	0x2133, 0x2134, 0x2135, 0x2136, 0x213c, 0x2147, 0x2149, 0x214b,
	0x214d, 0x214f, 0x2151, 0x2153, 0x2155, 0x2157, 0x2159, 0x216b,
	0x2242, 0x2244, 0
};

#define	COMPLEX_BGIMAGE \
    (bg_image_prev				\
     && ((bg_image_prev->ct_op == CTL_BIMAGE)	\
       || bg_image_prev->ct_op == CTL_BGRAD))

#define	POSY(size)	(-(int)((size)/2))

static void process_direc __P((struct render_state *, int *));

static int set_position __P((struct render_state *));
static void draw_line_start __P((struct render_state *));
static void draw_line_itemsize __P((struct render_state *, u_int));
static void draw_line_output __P((struct render_state *, u_char *));
static void draw_line_end __P((struct render_state *));
static void cutin __P((struct render_state *, int, int, int));
#if 0
static void shrink __P((u_char *, u_int));
#endif

static void draw_string __P((struct render_state *, u_char *));
static void draw_fragment __P((struct render_state *, u_char *, u_int, int));
static int iskinsokuchar __P((u_short));
static struct render_object *obj_alloc __P((struct render_state *state));
static void obj_free __P((struct render_state *, struct render_object *));
static int obj_new_xfont __P((struct render_state *, int, int, u_int));
static int obj_new_image __P((struct render_state *, int, int, Image *, int, int));
static int obj_new_icon __P((struct render_state *, int, int, u_int, u_int, u_long, u_int, XPoint *));
static Pixel obj_image_color __P((Image *, Image *, Pixel));
static void obj_draw_image __P((Drawable, u_int, u_int, struct render_object *));
static void obj_draw __P((struct render_state *, Drawable, u_int, u_int));
#ifdef VFLIB
static int obj_new_vfont __P((struct render_state *, int, int, struct vfont *));
static u_int draw_onechar_vf __P((struct render_state *, u_int, int, int, u_int, u_int));
#endif
static XFontStruct *x_setfont __P((u_char *, u_int, int));
static u_int draw_onechar_x __P((struct render_state *, u_int, int, int));

static void back_gradation __P((struct render_state *, struct ctrl_grad *));
static void image_load __P((struct render_state *, char *, int, int, int, int, int, int));
static void process_icon __P((struct render_state *, struct ctrl *));
static void draw_bar __P((struct render_state *, struct ctrl *));
static void process_system __P((struct render_state *, struct ctrl *));
static void process_xsystem __P((struct render_state *, struct ctrl *));
static Window search_child_window __P(());
static void reparent_child_window __P((Window, int, int));

/*
 * state management.
 */
void
state_goto(state, page, repaint)
	struct render_state *state;
	u_int page;
	int repaint;
{
	if (!repaint)
		purgechild(state->page);

	free_alloc_colors(&image_clr);
	free_alloc_colors(&font_clr);
	state->page = page;
	state->line = 0;
	state->cp = NULL;
	state->phase = P_NONE;
}

void
state_next(state)
	struct render_state *state;
{
	switch (state->phase) {
	case P_NONE:
		fprintf(stderr, "internal error\n");
		break;
	case P_DEFAULT:
		if (state->cp)
			state->cp = state->cp->ct_next;
		if (!state->cp) {
			state->cp = page_control[state->page][state->line];
			state->phase = P_PAGE;
		}
		break;
	case P_PAGE:
		if (state->cp)
			state->cp = state->cp->ct_next;
		if (!state->cp) {
			state->line++;
			state->cp = NULL;
			state->phase = P_NONE;
			state_init(state);
		}
		break;

	case P_END:
		/*nothing*/
		break;
	}

	/* next page */
	if (page_attribute[state->page].pg_linenum < state->line) {
		purgechild(state->page);
		if (state->page < maxpage) {
			state->phase = P_NONE;
			state->page++;
			state->line = 0;
			state_newpage(state);
			state_init(state);
		} else
			state->phase = P_END;
	}
}

void
state_init(state)
	struct render_state *state;
{
	assert(state);

	if (state->phase == P_NONE || !state->cp) {
#if 0
		if (!(page_attribute[state->page].pg_flag & PGFLAG_NODEF)) {
			state->cp = default_control[state->line];
			state->phase = P_DEFAULT;
		} else
#endif
		{
			state->cp = page_control[state->page][state->line];
			state->phase = P_PAGE;
		}
	}
}

void
state_newpage(state)
	struct render_state *state;
{
	state->ypos = 0;
	state->have_mark = 0;
}

/*
 * page management.
 */
void
draw_page(state, lastcp)
	struct render_state *state;
	struct ctrl *lastcp;
{
	u_int end_line;
	int pause;

	assert(state);

	/* initialize the state, if required. */
	if (state->phase == P_NONE || !state->cp) {
		state_newpage(state);
		state_init(state);
	}

	end_line = page_attribute[state->page].pg_linenum;

	while (1) {
		switch (state->phase) {
		case P_NONE:
			fprintf(stderr, "internal error\n");
			exit(-1);
		case P_DEFAULT:
		case P_PAGE:
			pause = 0;
			if (state->cp)
				process_direc(state, &pause);
			if (lastcp && state->cp == lastcp)
				goto done;
			if (pause) {
				if (state->cp
				 && state->cp->ct_op == CTL_PAUSE
				 && state->cp->cti_value) {
					goto done;
				}
			}
			break;
		case P_END:
			goto done;
		}
#if 0
		XFlush(display);
#endif
		state_next(state);
	}
done:
	XFlush(display);
}

Bool
draw_one(state, e)
	struct render_state *state;
	XEvent *e;
{
	u_int end_line;
	int pause;

	assert(state);

	/* initialize the state, if required. */
	if (!state->cp || state->phase == P_NONE) {
		state_newpage(state);
		state_init(state);
	}

	end_line = page_attribute[state->page].pg_linenum;

	switch (state->phase) {
	case P_NONE:
		fprintf(stderr, "internal error\n");
		exit(-1);
	case P_DEFAULT:
	case P_PAGE:
		pause = 0;
		if (state->cp)
			process_direc(state, &pause);
		if (pause) {
			/* seen pause, always block and wait the next event */
			XNextEvent(display, e);
			if (2 <= parse_debug) {
				fprintf(stderr,
					"pause and ot X11 event type=%d\n",
					e->type);
			}
			XFlush(display);
			return True;
		}
		if (XCheckMaskEvent(display, xeventmask, e) == True) {
			/* we got some event in the queue*/
			if (2 <= parse_debug) {
				fprintf(stderr,
					"interrupted and "
					"got X11 event type=%d\n",
					e->type);
			}
			state_next(state);
			return True;
		}

		/* we got no event in the queue */
		state_next(state);
		return False;
	case P_END:
		XNextEvent(display, e);
		if (2 <= parse_debug) {
			fprintf(stderr,
				"waited and got X11 event type=%d\n", e->type);
		}
		XFlush(display);
		return True;
	}
	fprintf(stderr, "internal error\n");
	exit(-1);
}

static void
process_direc(state, seenpause)
	struct render_state *state;
	int *seenpause;
{
	struct ctrl *cp;

	if (seenpause)
		*seenpause = 0;
	cp = state->cp;

	if (2 <= parse_debug) {
		fprintf(stderr, "p%d/l%d: ", state->page, state->line);
		debug0(cp);
	}

	switch(cp->ct_op) {
	case CTL_SIZE:
		char_size = window_height * cp->cti_value / 100;
		break;

	case CTL_VGAP:
		vert_gap = cp->cti_value;
		break;

	case CTL_HGAP:
		horiz_gap = cp->cti_value;
		break;

	case CTL_GAP:
		vert_gap = horiz_gap = cp->cti_value;
		break;

	case CTL_QUALITY:
		if (!quality_flag)
			b_quality = cp->cti_value;
		break;

	case CTL_PAUSE:
		if (seenpause)
			*seenpause = 1;
		break;

	case CTL_AGAIN:
		if (state->have_mark)
			state->ypos = state->mark_ypos;
		state->have_mark = 0;
		break;

	case CTL_FORE:
		fore_color = cp->ctl_value;
		XSetForeground(display, gcfore, fore_color);
		break;

	case CTL_BACK:
		bg_image_ctl = cp;	/*update later*/
		break;

	case CTL_CCOLOR:
		ctrl_color = cp->ctl_value;
		break;

	case CTL_CENTER:
		state->align = AL_CENTER;
		break;

	case CTL_LEFT:
		state->align = AL_LEFT;
		break;

	case CTL_LEFTFILL:
		state->align = AL_LEFTFILL0;
		break;

	case CTL_RIGHT:
		state->align = AL_RIGHT;
		break;

	case CTL_CONT:
		break;

#ifdef VFLIB
	case CTL_VFONT:
		vfc_setfont(cp->ctc_value);
		break;
#endif /*VFLIB*/

	case CTL_XFONT:
		state->xfont = cp->ctc_value;
		break;

	case CTL_BAR:
		draw_bar(state, cp);
		break;

	case CTL_IMAGE:
		image_load(state, cp->ctm_fname, cp->ctm_numcolor,
			cp->ctm_ximagesize, cp->ctm_yimagesize, 0,
			cp->ctm_zoomflag, 0);
		break;

	case CTL_BIMAGE:
		if (mgp_flag & FL_BIMAGE)
			break;
		bg_image_ctl = cp;	/*update later*/
		break;

	case CTL_BGRAD:
		if (mgp_flag & FL_BIMAGE)
			break;
		bg_image_ctl = cp;	/*update later*/
		break;

	case CTL_LCUTIN:
		state->special = SP_LCUTIN;
		break;

	case CTL_RCUTIN:
		state->special = SP_RCUTIN;
		break;

	case CTL_SHRINK:
		state->special = SP_SHRINK;
		break;

	case CTL_PREFIX:
		state->curprefix = cp->ctc_value;
		break;

	case CTL_TABPREFIX:
		state->tabprefix = cp->ctc_value;
		break;

	case CTL_PREFIXPOS:
	    {
		u_char *p;

		p = (state->tabprefix) ? state->tabprefix : state->curprefix;
		if (!p)
			break;
		draw_line_output(state, p);
		break;
	    }

	case CTL_TEXT:
		if (!cp->ctc_value)
			break;
		if (state->align == AL_LEFTFILL0) {
			state->align = AL_LEFTFILL1;
			state->leftfillpos = state->linewidth;
		}
		draw_line_output(state, cp->ctc_value);
		break;

	case CTL_LINESTART:
		if (state->line == 0) {
			if (bg_image_ctl) {
				if (bg_image_prev
				 && ctlcmp(bg_image_ctl, bg_image_prev) == 0) {
					; /* same as the last time */
				} else if (bg_image_ctl->ct_op == CTL_BIMAGE) {
				    image_load(state, bg_image_ctl->ctm_fname,
					bg_image_ctl->ctm_numcolor,
					bg_image_ctl->ctm_ximagesize,
					bg_image_ctl->ctm_yimagesize, 1,
					bg_image_ctl->ctm_zoomflag, 0);
				} else if (bg_image_ctl->ct_op == CTL_BGRAD) {
					back_gradation(state,
						&bg_image_ctl->ct_val.ctrl_grad);
				} else if (bg_image_ctl->ct_op == CTL_BACK) {
					XSetWindowBackground(display,
						state->target,
						bg_image_ctl->ctl_value);
					back_color = bg_image_ctl->ctl_value;
				} else {
					fprintf(stderr, "internal error: "
						"bogus background\n");
					exit(1);
				}
			}
			XClearWindow(display, state->target);
			bg_image_prev = bg_image_ctl;
			bg_image_ctl = NULL;
			if (t_fin)
				timebar(state);
		}
		draw_line_start(state);
		break;

	case CTL_LINEEND:
		draw_line_end(state);
		/* reset single-line oriented state */
		state->tabprefix = NULL;
		state->special = 0;
		if (state->align == AL_LEFTFILL1) {
			state->align = AL_LEFTFILL0;
			state->leftfillpos = 0;
		}
		break;

	case CTL_MARK:
		state->have_mark = 1;
		state->mark_ypos = state->ypos;
		break;

	case CTL_SYSTEM:
		process_system(state, cp);
		break;

	case CTL_XSYSTEM:
		process_xsystem(state, cp);
		break;

	case CTL_ICON:
		process_icon(state, cp);
		break;

	default:
		break;
	}
}

/*
 * line management.
 */
static int
set_position(state)
	struct render_state *state;
{
	int x;

	x = 0;
	switch (state->align) {
	case AL_CENTER:
		x = (window_width - state->linewidth)/ 2;
		break;

	case AL_LEFT:
	case AL_LEFTFILL0:
	case AL_LEFTFILL1:
		x = 0;
		break;

	case AL_RIGHT:
		x = window_width - state->linewidth;
		break;
	}

	return x;
}

static void
draw_line_start(state)
	struct render_state *state;
{
	struct render_object *obj;

	state->lineheight = 0;
	state->linewidth = 0;
	while ((obj = state->obj))
		obj_free(state, obj);
}

static void
draw_line_itemsize(state, newsize)
	struct render_state *state;
	u_int newsize;
{
	if (newsize > state->lineheight)
		state->lineheight = newsize;
}


static void
draw_line_output(state, data)
	struct render_state *state;
	u_char *data;
{
	draw_line_itemsize(state, VERT_STEP(char_size));
	draw_string(state, data);
}

static void
draw_line_end(state)
	struct render_state *state;
{
	u_int xpos;

	xpos = set_position(state);

	/* process the special attribute. */
	switch (state->special) {
#if 0
	case SP_SHRINK:
		shrink(data, page, xpos);
		break;
#endif
	case SP_LCUTIN:
		cutin(state, (int)xpos, state->ypos, 1);
		break;
	case SP_RCUTIN:
		cutin(state, (int)xpos, state->ypos, -1);
		break;
	default:
		break;
	}
	if (state->obj) {
		obj_draw(state, state->target, xpos, state->ypos);
		while (state->obj)
			obj_free(state, state->obj);
	}

	state->ypos += state->lineheight;
}

static void
cutin(state, lx, ly, dir)
	struct render_state *state;
	int lx;
	int ly;
	int dir;
{
	u_int step, x;
	int i;
	int sx;
	Window cutinWin;

	if (state->repaint)
		return;

	sx = (0 < dir) ? 0 : window_width - state->linewidth;
	step = (lx - sx) / 20;
	if (!step)
		step = 1;

	cutinWin = XCreateSimpleWindow(display, state->target,
		sx, ly, state->linewidth, state->lineheight, 0, fore_color, back_color);
	XSetWindowBackgroundPixmap(display, cutinWin, None);
	XMapSubwindows(display, state->target);

	if (state->obj) {
		obj_draw(state, cutinWin, 0, 0);
	}
	XFlush(display);

	x = sx;
	for (i = 0; i < 20; i++) {	/*XXX*/
		XMoveWindow(display, cutinWin, x, ly);
		XFlush(display);
		usleep(CUTIN_DELAY);
		x += step;
	}

	XDestroyWindow(display, cutinWin);
}

#if 0
static void
shrink(data, page)
	u_char *data;
	u_int page;
{
	u_int min_csize = char_size;
	u_int max_csize = window_height / 4;
	u_int csize, i, x;
	u_int step = (max_csize - min_csize) / 3;

	if (!step)
		step = 1;

	if (state->align != AL_CENTER) {
		fprintf(stderr, "align is not center: \n");
		return;
	}

	csize = char_size;
	for (i = max_csize; i > min_csize; i -= step) {
		char_size = i;
		draw_string(state, data);
		x = (window_width - state->linewidth) / 2;
		XCopyArea(display, maskpix, state->target, gc,
			0, 0, state->linewidth, char_size, x, state->ypos);
		XCopyArea(display, pixmap, state->target, gcor,
			0, 0, state->linewidth, char_size, x, state->ypos);
		XFlush(display);
		usleep(SHRINK_DELAY);
		XFillRectangle(display, pixmap, gcall,
			0, 0, window_width, char_size);
		XFillRectangle(display, maskpix, gcall,
			0, 0, window_width, char_size);
		XClearArea(display, state->target, x, state->ypos,
			state->linewidth, char_size, 0);
	}
	char_size = csize;
}
#endif

/*
 * render characters.
 */
static void
draw_string(state, data)
	struct render_state *state;
	u_char *data;
{
	u_char *p;
	u_char *q;
	u_int kanji = 0;
	u_int code2;

	p = data;

	while (*p && *p != '\n') {
		if (*p == 0x1b && *(p+1) == '$' &&
				(*(p+2) == 'B' || *(p+2) == '@')) {
			kanji = 1;
			p += 3;
			continue;
		}
		if (*p == 0x1b && *(p+1) == '(' &&
				(*(p+2) == 'B' || *(p+2) == 'J')) {
			kanji = 0;
			p += 3;
			continue;
		}

		if (!kanji && isspace(p[0])) {
			p++;
			state->linewidth +=
				HORIZ_STEP(char_size, char_size / 2);
			continue;
		}

		if (kanji) {
			for (q = p + 2; 0x21 <= *q && *q <= 0x7e; q += 2) {
				code2 = q[0] * 256 + q[1];
				if (!iskinsokuchar(code2))
					break;
			}
		} else {
			q = p;
			while (*q && isprint(*q) && !isspace(*q))
				q++;
			if (q == p)
				q++;
			else {
				/*
				 * append spaces to the end of the word.
				 * fragments in the following line:
				 *	"this is test"
				 * are:
				 *	"this_" "is_" "test"
				 */
				while (*q && isspace(*q))
					q++;
			}
		}

		draw_fragment(state, p, q - p, kanji);

		p = q;
	}
}

static void
draw_fragment(state, p, len, kanjimode)
	struct render_state *state;
	u_char *p;
	u_int len;
	int kanjimode;
{
	u_int char_len;
	u_short code;
	struct render_object *tail;
	struct render_object *thisline;
	struct render_object *thislineend;
	u_int startwidth;
	struct render_state backup;

	if (state->obj)
		tail = state->objlast;
	else
		tail = NULL;
	startwidth = state->linewidth;

	while (len) {
		code = kanjimode ? p[0] * 256 + p[1] : p[0];

		if (code == ' ') {
			char_len = char_size / 2;
			p++;
			len--;

			state->linewidth += HORIZ_STEP(char_size, char_len);
			continue;
		}

#ifdef USE_XDRAWSTRING
		char_len = draw_onechar_x(state, code, state->linewidth,
			POSY(char_size));
		if (char_len == 0 && (mgp_flag & FL_VERBOSE))
			printf("can't load font size %d "
				"(nor font in similar size)\n", char_size);
#else /*USE_XDRAWSTRING*/
# ifdef USE_XDRAWSTRING_ONLY_SMALL
		if (!kanjimode && char_size < 25) {
			char_len = draw_onechar_x(state, code,
				state->linewidth, POSY(char_size));
		} else
			char_len = 0;
		if (char_len == 0)
# endif /*USE_XDRAWSTRING_ONLY_SMALL*/
		{
			char_len = draw_onechar_vf(state, code,
				state->linewidth, POSY(char_size),
				kanjimode ? char_size
					  : (char_size * 4 / 5), /*XXX*/
				char_size);
		}
#endif /*USE_XDRAWSTRING*/

		p += (kanjimode ? 2 : 1);
		len -= (kanjimode ? 2 : 1);

		state->linewidth += HORIZ_STEP(char_size, char_len);
	}

	if (state->align == AL_LEFTFILL1
	 && window_width - state->leftfillpos / 2 < state->linewidth) {
		memcpy(&backup, state, sizeof(struct render_state));

		/* strip off the last fragment we wrote. */
		if (tail) {
			thisline = tail->next;
			thislineend = state->objlast;
			tail->next = NULL;
			state->objlast = tail;
		} else {
			thisline = state->obj;
			thislineend = state->objlast;
			state->obj = state->objlast = NULL;
		}
		draw_line_end(state);	/* flush the line. */

		/* start the new line with the last fragment we wrote. */
		draw_line_start(state);
		draw_line_itemsize(state, VERT_STEP(char_size));
		state->linewidth = state->leftfillpos;
		state->linewidth += (backup.linewidth - startwidth);
		if (state->obj && state->objlast)
			state->objlast->next = thisline;
		else
			state->obj = thisline;
		state->objlast = thislineend;

		/* fix up x position. */
		for (tail = state->obj; tail; tail = tail->next) {
			tail->x -= startwidth;
			tail->x += state->leftfillpos;
		}
	}
}

static int
iskinsokuchar(code)
	u_short code;
{
	u_short *kinsoku;

	for (kinsoku = kinsokutable; *kinsoku; kinsoku++) {
		if (code == *kinsoku)
			return 1;
	}
	return 0;
}

static struct render_object *
obj_alloc(state)
	struct render_state *state;
{
	struct render_object *obj;

	obj = malloc(sizeof(*obj));
	if (obj == NULL)
		return NULL;
	obj->next = NULL;
	if (state->obj == NULL)
		state->obj = obj;
	else
		state->objlast->next = obj;
	state->objlast = obj;
	return obj;
}

static void
obj_free(state, obj)
	struct render_state *state;
	struct render_object *obj;
{
	struct render_object *o;

	if (state->obj == obj)
		state->obj = obj->next;
	else {
		for (o = state->obj; o; o = o->next)
			if (o->next == obj)
				break;
		/* ASSERT(o != NULL); */
		o->next = obj->next;
	}
	if (state->objlast == obj)
		state->objlast = obj->next;
	switch (obj->type) {
#ifdef VFLIB
	case O_VFONT:
		obj->data.vfc->ref--;
		break;
#endif /* VFLIB */
	case O_IMAGE:
		freeImage(obj->data.image.image);
		break;
	case O_XFONT:
		/* do nothing */
		break;
	case O_ICON:
		if (obj->data.icon.xpoint)
			free(obj->data.icon.xpoint);
		break;
	}
	free(obj);
}

#ifdef VFLIB
static int
obj_new_vfont(state, x, y, vfc)
	struct render_state *state;
	int x, y;
	struct vfont *vfc;
{
	struct render_object *obj;

	obj = obj_alloc(state);
	if (obj == NULL)
		return 0;
	obj->x = x;
	obj->y = y;
	obj->fore = fore_color;
	obj->type = O_VFONT;
	obj->data.vfc = vfc;
	vfc->ref++;
	return 1;
}
#endif /* VFLIB */

static int
obj_new_xfont(state, x, y, code)
	struct render_state *state;
	int x, y;
	u_int code;
{
	struct render_object *obj;

	obj = obj_alloc(state);
	if (obj == NULL)
		return 0;
	obj->x = x;
	obj->y = y;
	obj->fore = fore_color;
	obj->type = O_XFONT;
	obj->data.xfont.xfont = state->xfont;
	obj->data.xfont.csize = char_size;
	obj->data.xfont.code = code;
	return 1;
}

static int
obj_new_image(state, x, y, image, xzoom, yzoom)
	struct render_state *state;
	int x, y;
	Image *image;
	int xzoom, yzoom;
{
	struct render_object *obj;

	obj = obj_alloc(state);
	if (obj == NULL)
		return 0;
	obj->x = x;
	obj->y = y;
	obj->type = O_IMAGE;
	obj->data.image.image = image;
	obj->data.image.xzoom = xzoom;
	obj->data.image.yzoom = yzoom;
	return 1;
}

static int
obj_new_icon(state, x, y, itype, isize, color, npoint, xpoint)
	struct render_state *state;
	int x, y;
	u_int itype, isize;
	u_long color;
	u_int npoint;
	XPoint *xpoint;
{
	struct render_object *obj;
	int i;

	obj = obj_alloc(state);
	if (obj == NULL)
		return 0;
	obj->x = x;
	obj->y = y;
	obj->fore = color;
	obj->type = O_ICON;
	obj->data.icon.itype = itype;
	obj->data.icon.isize = isize;
	obj->data.icon.npoint = npoint;
	if (npoint) {
		obj->data.icon.xpoint = malloc(sizeof(XPoint) * npoint);
		if (obj->data.icon.xpoint == NULL) {
			obj_free(state, obj);
			return NULL;
		}
		for (i = 0; i < npoint; i++)
			obj->data.icon.xpoint[i] = xpoint[i];
	} else
		obj->data.icon.xpoint = NULL;
	return 1;
}

static Pixel
obj_image_color(image, bimage, d)
	Image *image, *bimage;
	Pixel d;
{
	int i, j;
	RGBMap rgb;
	int r, g, b;
	u_int error, e;
	char hist[256];
	u_char *p;

	switch (bimage->type) {
	case IBITMAP:
		r = g = b = d ? 0xffff : 0;
		break;
	case IRGB:
		r = bimage->rgb.red[d];
		g = bimage->rgb.green[d];
		b = bimage->rgb.blue[d];
		break;
	case ITRUE:
		r = TRUE_RED(d) << 8;
		g = TRUE_GREEN(d) << 8;
		b = TRUE_BLUE(d) << 8;
		break;
	default:
		return 0;
	}
	for (i = 0; i < image->rgb.used; i++) {
		if (image->rgb.red[i] == r &&
		    image->rgb.green[i] == g &&
		    image->rgb.blue[i] == b)
			return i;
	}
	if (i >= image->rgb.size) {
		if (i >= 256) {
			/* search free slot */
			if (image->rgb.size == 256) {
				memset(hist, 0, sizeof(hist));
				p = image->data;
				for (j = 0; j < image->height; j++)
					for (i = 0; i < image->width; i++)
						hist[*p++] = 1;
				for (i = 0; i < 256; i++)
					if (hist[i] == 0)
						goto freeslot;
			}
			/* search nearest color */
			error = (u_int)-1;
			j = -1;
#define	DIFF(x, y)	(x > y ? x - y : y - x)
			for (i = 0; i < image->rgb.used; i++) {
				e = DIFF(image->rgb.red[i], r) +
				    DIFF(image->rgb.green[i], g) +
				    DIFF(image->rgb.blue[i], b);
				if (e < error) {
					j = i;
					error = e;
				}
			}
#undef DIFF
			if (error > 0x800)
				return -1;
			return j;
		}
		image->depth = 8;
		newRGBMapData(&rgb, depthToColors(image->depth));
		for (i = 0; i < image->rgb.used; i++) {
			rgb.red[i] = image->rgb.red[i];
			rgb.green[i] = image->rgb.green[i];
			rgb.blue[i] = image->rgb.blue[i];
		}
		rgb.used = i;
		freeRGBMapData(&image->rgb);
		image->rgb = rgb;
	}
  freeslot:
	image->rgb.red[i] = r;
	image->rgb.green[i] = g;
	image->rgb.blue[i] = b;
	if (image->rgb.used < i + 1)
		image->rgb.used = i + 1;
	return i;
}

static void
obj_draw_image(target, x, y, obj)
	Drawable target;
	u_int x, y;
	struct render_object *obj;
{
	Image *image, *timage;
	XImageInfo *ximageinfo;
	XImage *xim;
	int i, j;
	byte *p, *b;
	Pixel d, n, pd, pn;
	int pl, bpl;
	int verbose = mgp_flag & FL_VERBOSE;
	int private = mgp_flag & FL_PRIVATE;

	image = obj->data.image.image;
	if (obj->data.image.xzoom != 100 || obj->data.image.yzoom != 100) {
		timage = image;
		image = zoom(image, obj->data.image.xzoom, obj->data.image.yzoom, verbose);
		freeImage(timage);
	}
	if (COMPLEX_BGIMAGE && image->trans >= 0 && image->type == IRGB) {
		p = image->data;
		pl = image->pixlen;
		bpl = bg_image->pixlen;
		pd = pn = -1;
		for (j = 0; j < image->height; j++) {
  again:
			b = bg_image->data + (bg_image->width * (y + j) + x) * bpl;
			for (i = 0; i < image->width; i++) {
				d = memToVal(p, pl);
				if (d == image->trans) {
					d = memToVal(b, bpl);
					if (d == pd) {
						n = pn;
					} else {
						pd = d;
						pn = n = obj_image_color(image,
								bg_image, d);
						if (n == -1) {
							timage = image;
							image = reduce(image, 128, verbose);
							if (timage != image)
								freeImage(timage);
							pl = image->pixlen;
							p = image->data + image->width * j * pl;
							pd = pn = -1;

							goto again;
						}
					}
					valToMem(n, p, pl);
				}
				p += pl;
				b += bpl;
			}
		}
		if (verbose)
			printf("obj_draw_image: trans=%d, rgb_used=%d, rgb_size=%d\n", image->trans, image->rgb.used, image->rgb.size);
	}
	obj->data.image.image = image;	/* to free later */
	ximageinfo= imageToXImage(display, screen, visual, depth, image, private, 0,0, verbose);
	if (ximageinfo == NULL) {
		fprintf(stderr, "Cannot convert Image to XImage\n");
		exit(1);
	}
	xim = ximageinfo->ximage;
	XPutImage(display, target, gcfore, xim, 0, 0,
		x, y, xim->width, xim->height);
	freeXImage(image, ximageinfo);
}

static void
obj_draw(state, target, xpos, ypos)
	struct render_state *state;
	Drawable target;
	u_int xpos, ypos;
{
	struct render_object *obj;
	XImage *bim;
	u_int x, y;
	int width, height;
	u_long fore;
	u_int code;
	int kmode;
	XChar2b kch[2];
	u_char ch[2];
	u_int isize;
	int i;
	int offx, offy;
#ifdef VFLIB
	u_int charlen;
	XImage *xim;
	struct vfont *vfc;
	int hasvfont;
	u_long bcolor;
#endif /* VFLIB */

	/*
	 * very complicated...
	 *
	 *	xpos, ypos	x/y position within target.
	 *	offx, offy	x/y offset for object to target.
	 *	state->ypos	absolute y position in main window.
	 */
	offx = xpos;
	offy = ypos + state->lineheight/2;
	bim = COMPLEX_BGIMAGE ? bg_ximageinfo->ximage : NULL;
	width = (state->linewidth <= window_width - xpos)
			? state->linewidth
			: window_width - xpos;
	height = state->lineheight;
	fore = fore_color;

#ifdef VFLIB
	bcolor = back_color;
	if (bim) {
		hasvfont = 0;
		for (obj = state->obj; obj; obj = obj->next) {
			if (obj->type == O_VFONT) {
				hasvfont = 1;
				break;
			}
		}
		bcolor = 0;
		if (hasvfont &&
		    bim->width >= xpos + width &&
		    bim->height >= state->ypos + height) {
			bcolor = -1;
			xim = XSubImage(bim, xpos, state->ypos, width, height);
			for (obj = state->obj; obj; obj = obj->next) {
				if (obj->type == O_VFONT) {
					(void)vfc_image(obj->data.vfc,
						obj->fore, -1, xim,
						obj->x, obj->y + height / 2);
				}
			}
			XPutImage(display, target, gcfore, xim, 0, 0,
				xpos, ypos, width, height);
			XDestroyImage(xim);
		}
	}
#endif /* VFLIB */
	for (obj = state->obj; obj; obj = obj->next) {
		x = obj->x + offx;
		y = obj->y + offy;
		switch (obj->type) {
#ifdef VFLIB
		case O_VFONT:
			if (bcolor == -1)
				break;
			vfc = obj->data.vfc;
			xim = vfc_image(vfc, obj->fore, bcolor, bim, x, y);
			if (!xim)
				xim = vfc->image;
			charlen = vfc->charlen;
			if (vfc->minx + charlen > vfc->width)
				charlen = vfc->width - vfc->minx;
			XPutImage(display, target, gcfore, xim, vfc->minx, 0,
				x, y, charlen, vfc->height);
			if (xim != vfc->image)
				XDestroyImage(xim);
			break;
#endif /* VFLIB */
		case O_IMAGE:
			obj_draw_image(target, x, y, obj);
			break;
		case O_XFONT:
			code = obj->data.xfont.code;
			kmode = code & 0xff00 ? 1 : 0;
			(void)x_setfont(obj->data.xfont.xfont, obj->data.xfont.csize, kmode);
			if (obj->fore != fore) {
				fore = obj->fore;
				XSetForeground(display, gcfore, fore);
			}
			if (kmode) {
				kch[0].byte1 = (code >> 8) & 0xff;
				kch[0].byte2 = code & 0xff;
				XDrawString16(display, target, gcfore,
						x, y, kch, 1);
			} else {
				ch[0] = code & 0xff;
				XDrawString(display, target, gcfore,
						x, y, ch, 1);
			}
			break;
		case O_ICON:
			if (obj->fore != fore) {
				fore = obj->fore;
				XSetForeground(display, gcfore, fore);
			}
			isize = obj->data.icon.isize;
			switch (obj->data.icon.itype) {
			case 1: /* this is box */
				XFillRectangle(display, target, gcfore, x, y,
					isize, isize);
				break;
			case 2: /* this is arc */
				XFillArc(display, target, gcfore, x, y,
					isize, isize, 0, 360 * 64);
				break;
			case 3: case 4: case 5: case 6:
			case 7:
				for (i = 0; i < obj->data.icon.npoint; i++) {
					obj->data.icon.xpoint[i].x += offx;
					obj->data.icon.xpoint[i].y += offy;
				}
				XFillPolygon(display, target, gcfore, 
					obj->data.icon.xpoint,
					obj->data.icon.npoint,
					Convex, CoordModeOrigin);
				break;
			}
			break;
		}
	}
	if (fore != fore_color)
		XSetForeground(display, gcfore, fore_color);
	/* ASSERT(state->obj == NULL); */
	/* ASSERT(state->objlast == NULL); */
}

#ifdef VFLIB
static u_int
draw_onechar_vf(state, code, x, y, width, height)
	struct render_state *state;
	u_int code;
	int x, y;
	u_int width, height;
{
	struct vfont *vfc;

	vfc = vfc_get(code, width, height, 1);

	obj_new_vfont(state, x, y, vfc);
	return vfc->charlen;
}
#endif /* VFLIB */

static XFontStruct *
x_setfont(xfont, csize, kmode)
	u_char *xfont;
	u_int csize;
	int kmode;
{
	static u_int last_csize = 0;
	static u_char *last_xfont = NULL;
	static int last_kmode = -1;
	static XFontStruct *xfontstruct;
	int i, fsize;
	u_char fontstring[BUFSIZ];
#ifdef VFLIB
#define	FONTTYPEMAX	10	/* number of used font type (in cache) */
#define	FONTLISTMAX	20	/* number of list for specified font type */
#define	FONTALLOWMAX	105	/* % of desired font */
#define	FONTALLOWMIN	90	/* % of desired font */
	char **fontlist, **font, *p;
	u_int error;
	int best, freeindex, count;
	int maxsize, minsize;
	static struct {
		u_char *xfont;
		char **list;
		int count;
	} fonts[FONTTYPEMAX];
#endif /* VFLIB */

	if (xfont == last_xfont && csize == last_csize && kmode == last_kmode)
		return xfontstruct;

	/* load new font */
	if (xfontstruct)
		XFreeFont(display, xfontstruct);

	last_xfont = xfont;
	last_csize = csize;
	last_kmode = kmode;

	if (csize < 5) {
		xfontstruct = XLoadQueryFont(display, "nil2");
		goto gotfont;
	}
	/*
	 * Look for the best font possible.
	 * 1. Check for a font that is smaller than the required one.
	 *    By using smaller font, we won't make the screen garbled.
	 * 2. If 1. is impossible, look for slightly larger font than
	 *    the required one.
	 */
#ifdef VFLIB
	fontlist = NULL;
	freeindex = -1;
	for (i = 0; i < FONTTYPEMAX; i++) {
		if (fonts[i].xfont == NULL) {
			if (freeindex < 0)
				freeindex = i;
			continue;
		}
		if (fonts[i].xfont == xfont ||
		    strcmp(fonts[i].xfont, xfont) == 0) {
			fontlist = fonts[i].list;
			count = fonts[i].count;
			freeindex = i;
			break;
		}
	}
	if (fontlist == NULL) {
		if (kmode)
			sprintf(fontstring, KANJIFONT_FORMAT, 0);
		else
			sprintf(fontstring, FONT_FORMAT, xfont, 0);
		p = strchr(fontstring, '0');	/* != NULL */
		*p = '*';
		fontlist = XListFonts(display, fontstring, FONTLISTMAX, &count);
		if (fontlist == NULL)
			return NULL;
		if (freeindex >= 0) {
			fonts[freeindex].xfont = xfont;
			fonts[freeindex].list = fontlist;
			fonts[freeindex].count = count;
		}
	}
	error = (u_int)-1;
	best = -1;
	maxsize = csize * FONTALLOWMAX / 100;		/* truncate */
	minsize = (csize * FONTALLOWMIN + 99) / 100;	/* roundup */
	for (i = 0, font = fontlist; i < count; i++, font++) {
		for (p = *font; *p; p++)
			if (*p >= '0' && *p <= '9')
				break;
		if (!*p)
			continue;
		fsize = atoi(p);
		if (fsize == 0 || fsize > maxsize || fsize < minsize)
			continue;
		if (fsize > csize)
			fsize = fsize - csize + 100;
					/* penalty for larger font */
		else
			fsize = csize - fsize;
		if (error > fsize) {
			error = fsize;
			best = i;
		}
	}
	if (best >= 0)
		xfontstruct = XLoadQueryFont(display, fontlist[best]);
	else
		xfontstruct = NULL;
	if (freeindex < 0)
		XFreeFontNames(fontlist);
#else /* VFLIB */
#define	FONTSEARCH	20
	for (i = 0; i < FONTSEARCH; i++) {
		fsize = csize - i;
		if (fsize <= 0)
			fsize = i + 1;
		if (kmode) {
			sprintf(fontstring, KANJIFONT_FORMAT, fsize);
		} else {
			sprintf(fontstring, FONT_FORMAT, xfont, fsize);
		}
		if ((xfontstruct = XLoadQueryFont(display, fontstring)))
			break;
	}
#undef FONTSEARCH
#endif /* VFLIB */
  gotfont:
	if (xfontstruct == NULL)
		return NULL;
	XSetFont(display, gcfore, xfontstruct->fid);
	return xfontstruct;
}

static u_int
draw_onechar_x(state, code, x, y)
	struct render_state *state;
	u_int code;
	int x, y;
{
	u_int char_len;
	static XFontStruct *xfontstruct;
	int kanjimode;

	kanjimode = (code & 0xff00) ? 1 : 0;
	xfontstruct = x_setfont(state->xfont, char_size, kanjimode);
	if (xfontstruct == NULL)
		return 0;

#define YCONTROL	(xfontstruct->descent / 2)
	draw_line_itemsize(state,
		(char_size > xfontstruct->ascent)
			? char_size + xfontstruct->descent + 4
			: xfontstruct->ascent + xfontstruct->descent + 4);
	if (kanjimode) {
		XChar2b kch[2];

		kch[0].byte1 = (code >> 8) & 0xff;
		kch[0].byte2 = code & 0xff;

		char_len = XTextWidth16(xfontstruct, kch, 1);

		obj_new_xfont(state, x, y + char_size - YCONTROL, code);
	} else {
		u_char ch[2];
		ch[0] = code & 0xff;

		char_len = XTextWidth(xfontstruct, ch, 1);

		obj_new_xfont(state, x, y + char_size - YCONTROL, code);
	}
#undef YCONTROL

	return char_len;
}

/*
 * render misc items.
 */
static void
back_gradation(state, cg0)
	struct render_state *state;
	struct ctrl_grad *cg0;
{
	struct ctrl_grad cg1;
	struct ctrl_grad *cg;
	int srcwidth, srcheight;
	int dstwidth, dstheight;
	int dir, numcolor;
	int xzoomrate, yzoomrate;
	int hquality, vquality;

	Image *myimage, *image;
	Pixmap mypixmap;
	XImageInfo *ximageinfo;
	byte *pic;
	int verbose = mgp_flag & FL_VERBOSE;
	int private = mgp_flag & FL_PRIVATE;
	static Cursor curs;

	/* okay, please wait for a while... */
	if (!curs)
		curs = XCreateFontCursor(display, XC_watch);
	XDefineCursor(display, state->target, curs);
	XFlush(display);

	/* just for safety */
	memcpy(&cg1, cg0, sizeof(struct ctrl_grad));
	cg = &cg1;

	/* grab parameters */
	dir = cg->ct_direction;
	numcolor = cg->ct_numcolor;
	hquality = b_quality;
	vquality = b_quality;

	/*
	 * XXX zoomflag is too complex to understand.
	 */
	if (!cg->ct_zoomflag) {
		int t;
		int i;

		dstwidth = window_width * cg->ct_width / 100;
		dstheight = window_height * cg->ct_height / 100;
		srcwidth = dstwidth;
		srcheight = dstheight;

		/*
		 * apply quality factor if srcwidth/height are large enough.
		 */
#define TOOSMALLFACTOR 8
		t = srcwidth;
		for (i = 100; hquality < i; i--) {
			t = srcwidth * i / 100;
			if (t < cg->ct_g_colors * TOOSMALLFACTOR)
				break;
		}
		srcwidth = t;

		t = srcheight;
		for (i = 100; vquality < i; i--) {
			t = srcheight * i / 100;
			if (t < cg->ct_g_colors * TOOSMALLFACTOR)
				break;
		}
		srcheight = t;
#undef TOOSMALLFACTOR
	} else {
		dstwidth = window_width;
		dstheight = window_height;
		srcwidth = window_width * cg->ct_width / 100;
		srcheight = window_height * cg->ct_height / 100;

		/*
		 * we don't apply quality factor here, since srcwidth/height
		 * is already smaller than dstwidth/height.
		 */
	}

#if 0
	if (srcwidth * hquality / 100 < cg->ct_g_colors * TOOSMALLFACTOR
	 || srcheight * vquality / 100 < cg->ct_g_colors * TOOSMALLFACTOR) {
		srcwidth = srcwidth * hquality / 100;
		srcheight = srcheight * vquality / 100;
	}
#endif

	xzoomrate = 100 * dstwidth / srcwidth;
	yzoomrate = 100 * dstheight / srcheight;

	/* performace enhance hack for special case */
	if (dir % 90 == 0) {
		int *p;
		int *q;
		int primes[] = { 2, 3, 5, 7, 11, 13, 0 };
		int i;

		/*
		 * 0 or 180: reduce width
		 * 90 or 270: reduce height
		 */
		p = (dir % 180 == 0) ? &srcwidth : &srcheight;
		q = (dir % 180 == 0) ? &xzoomrate : &yzoomrate;

		for (i = 0; primes[i]; i++) {
			while (*p % primes[i] == 0) {
				*p = *p / primes[i];
				*q = *q * primes[i];
			}
		}
	}

	if (verbose) {
		fprintf(stderr, "raw: %d,%d qu: %d,%d "
			"dst: %d,%d src: %d,%d zoom: %d,%d\n",
			cg->ct_width, cg->ct_height,
			hquality, vquality,
			dstwidth, dstheight, srcwidth, srcheight,
			xzoomrate, yzoomrate);
	}

	screen = DefaultScreen(display);

	/* make gradation image */
	pic = draw_gradation(srcwidth, srcheight, cg);
	myimage = make_XImage(pic, srcwidth, srcheight);

	if (numcolor < 64)
		myimage = reduce(myimage, numcolor, verbose);

#ifdef DEBUG
	fprintf(stderr, "Draw:zoomrate:(%d,%d)->(%d,%d)\n",
		cg->ct_xzoom,cg->ct_yzoom,xzoomrate,yzoomrate);
	fprintf(stderr, "zoom mode %d: (%d, %d)->(%d, %d)[%d]\n",
		zoomflag, width, height, window_width, window_height,
		b_quality);
#endif

	if (xzoomrate != 100 || yzoomrate != 100) {
		image = myimage;
		myimage = zoom(image, xzoomrate, yzoomrate, verbose);
		freeImage(image);
	}

	if (private) free_alloc_colors(&back_clr);
	ximageinfo = imageToXImage(display, screen, visual, depth, myimage,
		private, 0, 1, verbose);
	if (!ximageinfo) {
		fprintf(stderr, "Cannot convert Image to XImage\n");
		exit(1);
	}

	mypixmap = ximageToPixmap(display, DefaultRootWindow(display),
		ximageinfo);
	if (mypixmap == None && verbose) {
		fprintf(stderr, "Cannot create image in server\n");
		exit(1);
	}

	/* draw background */
	XSetWindowBackgroundPixmap(display, state->target, mypixmap);

	/* finish */
	XFreePixmap(display, mypixmap);
	if (bg_image) {
		freeXImage(bg_image, bg_ximageinfo);
		freeImage(bg_image);
	}
	bg_ximageinfo = ximageinfo;
	bg_image = myimage;
	XUndefineCursor(display, state->target);
	XFlush(display);
}

static void
image_load(state, filename, numcolor, ximagesize, yimagesize, backflag, zoomflag, centerflag)
	struct render_state *state;
	char *filename;
	int numcolor;
	int ximagesize;
	int yimagesize;
	int backflag;
	int zoomflag;
	int centerflag;
{
	Image *image, *myimage;
	Pixmap mypixmap;
	XImageInfo *ximageinfo;
	u_int image_posx;
	int width, height;
	int xzoomrate, yzoomrate;
	int	verbose = mgp_flag & FL_VERBOSE;
	int	private = mgp_flag & FL_PRIVATE;
	static Cursor curs;
	static char backfile[MAXPATHLEN];
	static int backzoom, backnumcolor, backx, backy;

	if (!curs)
		curs = XCreateFontCursor(display, XC_watch);
	XDefineCursor(display, state->target, curs);
	XFlush(display);

	if ((myimage = loadImage(filename, verbose)) == NULL) {
		fprintf(stderr, "failed to load image file\n");
		exit(-1);
	}
	width = myimage->width;
	height = myimage->height;

	if (myimage->depth == 1) {
		XColor xc;

		xc.flags = DoRed | DoGreen | DoBlue;
		xc.pixel = fore_color;
		XQueryColor(display, colormap, &xc);
		*(myimage->rgb.red + 1) = xc.red;
		*(myimage->rgb.green + 1) = xc.green;
		*(myimage->rgb.blue + 1) = xc.blue;
		xc.pixel = back_color;
		XQueryColor(display, colormap, &xc);
		*myimage->rgb.red = xc.red;
		*myimage->rgb.green = xc.green;
		*myimage->rgb.blue = xc.blue;
	}

	if (numcolor)
		myimage = reduce(myimage, numcolor, verbose);

	if (zoomflag == 2) {
		/*
		 * auto resize according to physical and desired screen size.
		 * allow 5% error for '-o' option.
		 */
		if (ximagesize == 0 || ximagesize == window_width)
			ximagesize = 100;
		else
			ximagesize = window_width * 100 / ximagesize;
		if (yimagesize == 0 || yimagesize == window_height)
			yimagesize = 100;
		else
			yimagesize = window_height * 100 / yimagesize;
		if (ximagesize > 95 && ximagesize < 105 &&
		    yimagesize > 95 && yimagesize < 105)
			ximagesize = yimagesize = 0;
	}
	if (ximagesize != 0 && yimagesize != 0) {
		if (!zoomflag) {
			xzoomrate = window_width * ximagesize / width;
			yzoomrate = window_height * yimagesize / height;
		} else {
			xzoomrate = ximagesize;
			yzoomrate = yimagesize;
		}
	} else
		xzoomrate = yzoomrate = 100;

	if (backflag) {
		if (xzoomrate != 100 || yzoomrate != 100) {
			image = myimage;
			myimage = zoom(image, xzoomrate, yzoomrate, verbose);
			freeImage(image);
		}

		if (private) free_alloc_colors(&back_clr);
		ximageinfo= imageToXImage(display, screen, visual, depth,
				myimage, private, 0, 1, verbose);
		if (ximageinfo == NULL) {
			fprintf(stderr, "Cannot convert Image to XImage\n");
			exit(1);
		}

		mypixmap = ximageToPixmap(display,
				RootWindow(display, screen), ximageinfo);
		if (mypixmap == None && verbose) {
			fprintf(stderr, "Cannot create image in server\n");
			exit(1);
		}
		XSetWindowBackgroundPixmap(display, state->target, mypixmap);
		XFreePixmap(display, mypixmap);
		strcpy(backfile, filename);
		backnumcolor = numcolor;
		backzoom = zoomflag;
		backx = ximagesize;
		backy = yimagesize;
		if (bg_image) {
			freeXImage(bg_image, bg_ximageinfo);
			freeImage(bg_image);
		}
		bg_ximageinfo = ximageinfo;
		bg_image = myimage;
		goto end;
	}

	draw_line_itemsize(state, height * yzoomrate / 100);
	if (centerflag){
		image_posx = char_size / 2 - (width * xzoomrate / 100) /2;
	} else
		image_posx = 0;

	obj_new_image(state, state->linewidth + image_posx,
		- height * yzoomrate / 100 / 2,
		myimage, xzoomrate, yzoomrate);
	state->linewidth += (width * xzoomrate / 100);
end:
	XUndefineCursor(display, state->target);
	XFlush(display);
}

void
timebar(state)
	struct render_state *state;
{
	int pos, n, p, barlen;
	GC *pgc;

	if (t_start == 0 || tbar_mode == 0)
		return;

	pos = (window_width - 2) * (state->page - 1) / (maxpage - 1);
	p = (time(NULL) - t_start) * 100;
	barlen = window_width - window_width * p / t_fin / 6000;

	if (window_width * 50 / 100 < barlen)
		pgc = &gcgreen;
	else if (window_width * 30 / 100 < barlen)
		pgc = &gcyellow;
	else
		pgc = &gcred;
	if (barlen > 0) {
		XClearArea(display, state->target, 0, window_height - 2,
			window_width, 2, 0);
		XFillRectangle(display, state->target, *pgc,
			window_width - barlen, window_height - 1, barlen, 1);
		XFillRectangle(display, state->target, *pgc,
			pos, window_height - 5, 2, 5);
	} else if (barlen < 0) {
		barlen = - barlen;
		n = p / t_fin / 6000;
		if (n > window_height - 1)
			n = window_height - 1;
		if (n)
			XFillRectangle(display, state->target, gcred,
				0, window_height - n,
				barlen, n);
		XClearArea(display, state->target, 0, window_height - (n + 2),
			window_width, n + 2, 0);
		XFillRectangle(display, state->target, gcred,
			0, window_height - (n + 1),
			barlen % window_width, n + 1);
		XFillRectangle(display, state->target, gcred,
			pos, window_height - (n + 1 + 4),
			2, 5);
	}
}

static void
process_icon(state, cp)
	struct render_state *state;
	struct ctrl *cp;
{
	u_int i, icon_type, icon_size, icon_x, icon_y, index;
	u_long tmp_color;
	static struct ctl_words icon_words[] = {
		{ 1, "box", 3 },
		{ 2, "arc", 3 },
		{ 3, "delta1", 6 },
		{ 4, "delta2", 6 },
		{ 5, "delta3", 6 },
		{ 6, "delta4", 6 },
		{ 7, "dia", 3 },
		{ 0, NULL, 0 }
	};
	XPoint xpoint[4];
	static struct icon_point {
		int	point_num;
		XPoint xpoint[4];
	} icon_point[] = {{ 3, {{50,  0}, {25, 50}, {75, 50}, { 0, 0}}},
			  { 3, {{25,  0}, {75,  0}, {50, 50}, { 0, 0}}},
			  { 3, {{25,  0}, {25, 50}, {75, 25}, { 0, 0}}},
			  { 3, {{25, 25}, {75,  0}, {75, 50}, { 0, 0}}},
			  { 4, {{50,  0}, {25, 25}, {50, 50}, {75, 25}}}};
		
	for (i = 0; icon_words[i].ctl_strlen != 0; i++) {
		if (!strncasecmp(cp->ctic_value, icon_words[i].ctl_string,
			strlen(cp->ctic_value))) {
				break;
		}
	}

	icon_type = icon_words[i].ctl_type; /* may be 0 */
	icon_size = char_size * cp->ctic_size / 100;

	switch(icon_type){
	case 0:
		/* this is image */
		icon_x = icon_size * 100 / window_width;
		icon_y = icon_size * 100 / window_height;
		if (icon_x == 0) icon_x = 1;
		if (icon_y == 0) icon_y = 1;
		tmp_color = fore_color;
		fore_color = cp->ctic_color;
		image_load(state, cp->ctic_value, 0, icon_x, icon_y, 0, 0, 1);
		fore_color = tmp_color;
		break;

	case 1:
		/* this is box */
		obj_new_icon(state,
			state->linewidth + char_size/2 - icon_size/2,
			POSY(icon_size), icon_type, icon_size,
			cp->ctic_color, 0, NULL);
		state->linewidth += char_size;
		break;

	case 2:
		/* this is arc */
		obj_new_icon(state,
			state->linewidth + char_size/2 - icon_size/2,
			POSY(icon_size), icon_type, icon_size, 
			cp->ctic_color, 0, NULL);
		state->linewidth += char_size;
		break;

	case 3:
	case 4:
	case 5:
	case 6:
	case 7:
		index = icon_type - 3;
		for (i = 0; i < icon_point[index].point_num; i ++){
			xpoint[i].x = state->linewidth +
				icon_point[index].xpoint[i].x * char_size / 100;
			xpoint[i].y = POSY(icon_size) +
				icon_point[index].xpoint[i].y * char_size / 100;
		}
		obj_new_icon(state, 0, 0, icon_type, icon_size, 
			cp->ctic_color, icon_point[index].point_num, xpoint);
		state->linewidth += char_size;
		break;

	default:
		break;
	}

	cp = NULL;
}

static void
draw_bar(state, cp)
	struct render_state *state;
	struct ctrl *cp;
{
	u_int width, swidth, st, len;
	XColor col, scol;
	static GC gcbar, gcsbar;
	static u_long prevcolor = -1;

	if (!gcbar) {
		gcbar = XCreateGC(display, state->target, 0, 0);
		XSetFunction(display, gcbar, GXcopy);
		gcsbar = XCreateGC(display, state->target, 0, 0);
		XSetFunction(display, gcsbar, GXcopy);
	}
	col.pixel = cp->ctb_color;
	if (col.pixel == -1)
		col.pixel = fore_color;
	if (col.pixel != prevcolor) {
		prevcolor = col.pixel;
		col.flags = DoRed|DoGreen|DoBlue;
		XQueryColor(display, colormap, &col);
		scol.red   = col.red   / 2;
		scol.green = col.green / 2;
		scol.blue  = col.blue  / 2;
		if (!XAllocColor(display, colormap, &scol))
			scol.pixel = col.pixel;
		XSetForeground(display, gcbar, col.pixel);
		XSetForeground(display, gcsbar, scol.pixel);
	}
	width = cp->ctb_width * window_height / 1000;
	swidth = width / 2;
	width -= swidth;
	st = cp->ctb_start * window_width / 100;
	len = cp->ctb_length * window_width / 100;
	XFillRectangle(display, state->target, gcbar, st, state->ypos, len, width);
	XFillRectangle(display, state->target, gcsbar, st, state->ypos + width, len, swidth);

	state->ypos += width + swidth + VERT_GAP(char_size) / 2;
}

static void
process_system(state, cp)
	struct render_state *state;
	struct ctrl *cp;
{
	pid_t pid;
	int i;
	char **argv;
	char buf[BUFSIZ];

	if (mgp_flag & FL_NOFORK) {
		if (mgp_flag & FL_VERBOSE) {
			fprintf(stderr, "WARN: %%system ");
			for (i = 0; i < cp->cta_argc; i++) {
				fprintf(stderr, "%c%s", (i == 0) ? '"' : ' ',
					cp->cta_argv[i]);
			}
			fprintf(stderr, "\": directive skipped\n");
		}
		return;
	}

	if (checkchild(cp) != (pid_t)-1)
		return;	/*already running*/

	/*
	 * edit argument.
	 * if we have X11 geometry string 
	 */
	argv = (char **)cp->cta_argv;
	for (i = 0; i < cp->cta_argc; i++) {
		if (*(argv[i]) == '%')
			break;
	}
	if (i < cp->cta_argc) {
		char *p;
		char *q;
		int myxpos, myypos;
		int rootxsiz, rootysiz;
		int xsiz, ysiz;
		int xloc, yloc;
		int mode;

	    {
		XWindowAttributes wa;
		Window junkwin;
		int junk;

		XGetWindowAttributes(display, window, &wa);
		XTranslateCoordinates(display, window, wa.root,
			-wa.border_width, -wa.border_width,
			&myxpos, &myypos, &junkwin);
		XGetGeometry(display, wa.root, &junkwin, &junk, &junk,
			&rootxsiz, &rootysiz, &junk, &junk);
	     }

		argv = (char **)malloc((cp->cta_argc + 1) * sizeof(char *));
		memcpy(argv, cp->cta_argv, (cp->cta_argc + 1) * sizeof(char *));
		p = argv[i];
		p++;	/*drop percent char*/
		q = buf;
		*q = '\0';

		mode = XParseGeometry(p, &xloc, &yloc, &xsiz, &ysiz);
		if (mode == 0)
			goto fail;
		if ((mode & WidthValue) && (mode & HeightValue)) {
			sprintf(q, "%dx%d", xsiz * window_width / 100,
				ysiz * window_height / 100);
			q += strlen(q);
		}
		if ((mode & XValue) && (mode & YValue)) {
			xloc = xloc * window_width / 100;
			yloc = yloc * window_height / 100;
			if (mode & XNegative)
				xloc = rootxsiz - myxpos + window_width - xloc;
			else
				xloc += myxpos;
			if (mode & YNegative)
				yloc = rootysiz - myypos + window_height - yloc;
			else
				yloc += myypos;
			sprintf(q, "+%d+%d", xloc, yloc);
		}

		if (mgp_flag & FL_VERBOSE) {
			fprintf(stderr, "relative geometry: "
				"%s (presentation %dx%d+%d+%d)\n",
				argv[i], window_width, window_height,
				myxpos, myypos);
			fprintf(stderr, "\t-> %s\n", buf);
		}
		argv[i] = buf;

		if (0) {
fail:
			if (mgp_flag & FL_VERBOSE) {
				fprintf(stderr,
					"relative geometry: %s failed\n",
					argv[i]);
			}
		}
	}
	pid = fork();
	if (pid < 0) {
		perror("fork");
		exit(1);
	} else if (pid == 0) {
		execvp(argv[0], argv);
		exit(1);
	}

	if (!cp->cta_flag)	/*will be purged at the end of page*/
		regchild(pid, cp, -1, state->page);
	else
		regchild(pid, cp, -1, cp->cta_flag);
}

static void
process_xsystem(state, cp)
	struct render_state *state;
	struct ctrl *cp;
{
	pid_t pid;
	int i, dumint;
	int xloc, yloc;
	int xsiz, ysiz;
	char **argv;
	char buf[BUFSIZ];
	Window window_id, dumwin;

	if (mgp_flag & FL_NOFORK) {
		if (mgp_flag & FL_VERBOSE) {
			fprintf(stderr, "WARN: %%system ");
			for (i = 0; i < cp->cta_argc; i++) {
				fprintf(stderr, "%c%s", (i == 0) ? '"' : ' ',
					cp->cta_argv[i]);
			}
			fprintf(stderr, "\": directive skipped\n");
		}
		return;
	}


	/*
	 * edit argument.
	 * if we have X11 geometry string 
	 */
	argv = (char **)cp->cta_argv;
	for (i = 0; i < cp->cta_argc; i++) {
		if (!strncmp(argv[i], "-geom", 5))
			break;
	}
	i ++;
	if (i < cp->cta_argc) {
		char *p;
		char *q;
		int mode;

		argv = (char **)malloc((cp->cta_argc + 1) * sizeof(char *));
		memcpy(argv, cp->cta_argv, (cp->cta_argc + 1) * sizeof(char *));
		p = argv[i];
		if (*p == '%') p++;	/*drop percent char*/
		q = buf;
		*q = '\0';

		mode = XParseGeometry(p, &xloc, &yloc, &xsiz, &ysiz);
		if (mode == 0)
			goto fail;
		if ((mode & WidthValue) && (mode & HeightValue)) {
			xsiz = xsiz * window_width / 100;
			ysiz = ysiz * window_height / 100;
			sprintf(q, "%dx%d", xsiz, ysiz);
			q += strlen(q);
		}
		/* make window raise outside of display */
		sprintf(q, "+%d+%d", DisplayWidth(display, DefaultScreen(display)),
						DisplayHeight(display, DefaultScreen(display)));

		if (mgp_flag & FL_VERBOSE) {
			fprintf(stderr, "relative geometry: "
				"%s (presentation %dx%d+%d+%d)\n",
				argv[i], window_width, window_height,
				xloc, yloc);
			fprintf(stderr, "\t-> %s\n", buf);
		}
		argv[i] = buf;

		if (0) {
fail:
			if (mgp_flag & FL_VERBOSE) {
				fprintf(stderr,
					"relative geometry: %s failed\n",
					argv[i]);
			}
		}
	} else {
		u_char	geom_arg1[] = {"-geometry"};
		u_char  geom_arg2[512];

		sprintf(geom_arg2, "+%d+%d", DisplayWidth(display, 
			DefaultScreen(display)),
			DisplayHeight(display, DefaultScreen(display)));

		argv[cp->cta_argc] = geom_arg1;
		argv[cp->cta_argc+1] = geom_arg2;
	}

#if 0
	xloc = state->linewidth;
#endif
	state->linewidth = xsiz;
	xloc = set_position(state);
	yloc = state->ypos;

	if ((window_id = checkchildwin(cp)) != (Window)-1)
		goto finish;	/*already running*/

	if (checkchild(cp) != (pid_t)-1)
		return;	/*already running*/

	pid = fork();
	if (pid < 0) {
		perror("fork");
		exit(1);
	} else if (pid == 0){
		usleep(EXEC_DELAY);
		execvp(argv[0], argv);
		exit(1);
	}

	window_id = search_child_window();

	if (!cp->cta_flag)	/*will be purged at the end of page*/
		regchild(pid, cp, window_id, state->page);
	else
		regchild(pid, cp, window_id, cp->cta_flag);

	if (window_id != -1)
		reparent_child_window(window_id, xloc, yloc);
	else {
		if (mgp_flag & FL_VERBOSE) {
			fprintf(stderr, "WARN: %%xsystem can not find child window:");
			for (i = 0; i < cp->cta_argc; i++) {
				fprintf(stderr, "%c%s", (i == 0) ? '"' : ' ',
					cp->cta_argv[i]);
			}
			fprintf(stderr, "\"\n");
		}
		return;
	}

#if 0
	state->linewidth += xsiz;
	state->lineheight += ysiz;
#endif

finish:
	XGetGeometry(display, window_id, &dumwin, 
		&xloc, &yloc, &xsiz, &ysiz, &dumint, &dumint);
	state->ypos += ysiz;
}


Window 
search_child_window()
{
	XEvent e;
	int	fd, found = 0;
	fd_set fdset, dumfdset; 
	struct timeval timeout;

	fd = ConnectionNumber(display);
	/* waiting for 2 second */
	timeout.tv_sec = 2;
	timeout.tv_usec = 0;

	/* get all client's ReparentNotify event */
	XSelectInput(display, DefaultRootWindow(display),
		SubstructureNotifyMask);

	while (!found) {
		while (XEventsQueued(display, QueuedAfterFlush) > 0) {
			XNextEvent(display, &e);
			if (e.type == ReparentNotify){
				found = 1;
				break;
			}
		}
		if (found) break;
		FD_ZERO(&fdset);
		FD_SET(fd, &fdset); 	
		FD_ZERO(&dumfdset);
		if (!select(fd+1, &fdset, &dumfdset, &dumfdset, &timeout))
			break;
	}

	XSelectInput(display, DefaultRootWindow(display), NoEventMask);

	if (found == 1)
		return e.xreparent.window;
	else
		return (Window)-1;
}

void 
reparent_child_window(child_window, x, y)
	Window	child_window;
	int	x,y;
{
	Window	dummyroot, *dummywin;
	Window	target, parent;
	u_int	dumint;

	target = child_window;
	while (1) {
		XQueryTree(display, target, &dummyroot, &parent, &dummywin,
			&dumint);
		if (parent == dummyroot)	
			break;
		XFree(dummywin);
		target = parent;
	}
	XReparentWindow(display, child_window, window, x, y);
	XDestroyWindow(display, target);
}
